#!/usr/bin/env python3
# -*- coding:utf-8; mode:python -*-
#
# Copyright 2020, 2021 Pradyumna Paranjape
# This file is part of psprint.
#
# psprint is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# psprint is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with psprint.  If not, see <https://www.gnu.org/licenses/>.
#
'''
tests
'''



#!/usr/bin/env python3
# -*- coding:utf-8; mode:python -*-
#
# Copyright 2020, 2021 Pradyumna Paranjape
# This file is part of psprint.
#
# psprint is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# psprint is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with psprint.  If not, see <https://www.gnu.org/licenses/>.
#
"""
Test configurations
"""
import os
import tempfile
import unittest
from configparser import ConfigParser
from pathlib import Path

import toml
import yaml
from psprint import errors, init_print

CONFIG_YML = '\n'.join([
    'FLAGS:',
    '  short: No',
    '  pad: Yes',
    '',
    'test:',
    '  pref: TEST',
    '  pref_s: t',
    '  pref_color: y',
    '  pref_bgcol: blue',
    '  pref_gloss: 1',
    '  text_color: k',
    '  text_bgcol: 7',
    '  text_gloss: d',
    '',
])


class TestPrintSpace(unittest.TestCase):
    def test_yml(self):
        '''
        yml configuration marks
        '''
        with tempfile.NamedTemporaryFile('wt', delete=False) as config:
            config.write(CONFIG_YML)
        test_print = init_print(Path(config.name))
        self.assertIn('test', test_print)

    def test_toml(self):
        '''
        yml configuration marks
        '''
        with tempfile.NamedTemporaryFile('wt', delete=False) as config:
            config_toml = toml.dumps(yaml.safe_load(CONFIG_YML))
            print(config_toml)
            config.write(config_toml)
        test_print = init_print(Path(config.name))
        self.assertIn('test', test_print)

    def test_bad_mark(self):
        '''
        bad mark defined
        '''
        with tempfile.NamedTemporaryFile('wt', delete=False) as config:
            bad_yml = CONFIG_YML.replace('test:', 'test: []')
            config.write(bad_yml)
        self.assertRaises(errors.BadMark, init_print, Path(config.name))

    def tearDown(self):
        if 'PSPRINTIC' in os.environ:
            del os.environ['PSPRINTRC']

    def setUp(self):
        if 'PSPRINTIC' in os.environ:
            del os.environ['PSPRINTRC']

    def test_no_env_var(self):
        os.environ['PSPRINTRC'] = 'BAD FILE NAME'
        self.assertRaises(FileNotFoundError, init_print)
        del os.environ['PSPRINTRC']

    def test_psprintrc(self):
        os.environ['PSPRINTRC'] = '../.psprintrc'
        init_print()



#!/usr/bin/env python3
# -*- coding:utf-8; mode:python -*-
#
# Copyright 2020, 2021 Pradyumna Paranjape
# This file is part of psprint.
#
# psprint is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# psprint is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with psprint.  If not, see <https://www.gnu.org/licenses/>.
#
"""
Test print iteration
"""

from unittest import TestCase

from psprint import errors, print


class TestIter(TestCase):
    """
    These cases will always pass
    Tester should check the output if it is manually appealing
    by calling KeyboardInterrupt

    """
    def test_dict(self):
        obj = {'1': 1, '2': 2, '3': 3}
        print(obj, mark='bug', pref='dict', end='\n', iterate=True)
        print(obj, mark='info', iterate=False)
        # raise KeyboardInterrupt

    def test_list(self):
        obj = [1, 2, 3]
        print(obj, mark='bug', pref='list', iterate=True)
        print(obj, mark='info', iterate=False)
        # raise KeyboardInterrupt

    def test_generator(self):
        obj = (k for k in (1, 2, 3))
        print(obj, mark='bug', pref='gen', iterate=True)
        print(obj, mark='info', iterate=False)
        # raise KeyboardInterrupt

    def test_Tuple(self):
        obj = (1, 2, 3)
        print(obj, mark='bug', pref='tuple', iterate=True)
        print(obj, mark='info', iterate=False)
        # raise KeyboardInterrupt

    def test_nest(self):
        obj = (k for k in ([], {
            'a': 1,
            'b': [2, 'b'],
            3: {
                'int': 3,
                'bool': bin(3),
                'hex': hex(3)
            }
        }, {1, 2, 3, 4}, [['a', 'A'], ['b', 'B'], ['c', 'C']]))
        print(obj, mark='bug', pref='nest', iterate=True)
        print(obj, mark='info', iterate=False)
        # raise KeyboardInterrupt

    def test_deep_stack(self):
        obj = [[[[[[[[[[[[[[[[[], [], []]]]]]]]]]]]]]]]]
        self.assertWarns(errors.DeepIterWarning,
                         print,
                         obj,
                         mark='bug',
                         pref='deep',
                         iterate=True)
        print(obj, mark='info', iterate=False)

    def test_long_stack(self):
        obj = range(1000)
        self.assertWarns(errors.LongIterWarning,
                         print,
                         obj,
                         mark='bug',
                         pref='long',
                         iterate=True)
        print(obj, mark='info', iterate=False)



#!/usr/bin/env python3
# -*- coding:utf-8; mode:python -*-
#
# Copyright 2020, 2021 Pradyumna Paranjape
# This file is part of psprint.
#
# psprint is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# psprint is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with psprint.  If not, see <https://www.gnu.org/licenses/>.
#
"""
Test psfmt
"""

import unittest

from psprint import DEFAULT_PRINT

DEFAULT_PRINT.switches['disabled'] = False
DEFAULT_PRINT.switches['short'] = False
DEFAULT_PRINT.switches['pad'] = False
DEFAULT_PRINT.switches['bland'] = True
psfmt = DEFAULT_PRINT.psfmt


class MyFmtClass():
    """My Test Class with format string"""
    def __init__(self):
        self.attr = 'data'

    def __repr__(self) -> str:
        return f'data: {self.attr!s}'

    def __format__(self, spec):
        fmt_out = []
        for line_no, line in enumerate(self.__repr__().split("\n")):
            if line_no == 0:
                fmt_out.extend(psfmt(line, mark=spec))
            else:
                fmt_out.extend(psfmt(line, mark='cont'))
        return '\n'.join(fmt_out)

    def update(self, data: str):
        self.attr = data


class TestFmtStr(unittest.TestCase):
    def setUp(self):
        self.fmtobj = MyFmtClass()

    def test_one_word(self):
        self.assertEqual(f'{self.fmtobj:list}', '[LIST]data: data')

    def test_one_line(self):
        data = 'The Quick Brown Fox Jumps Over The Lazy Dog'
        self.fmtobj.update(data)
        self.assertEqual(f'{self.fmtobj:list}', f'[LIST]data: {data}')

    def test_multi_line(self):
        data = ['The Quick Brown Fox', 'Jumps Over', 'The Lazy Dog']
        expected_out_str = '[LIST]data: ' + '\n'.join(data)
        self.fmtobj.update('\n'.join(data))
        self.assertEqual(f'{self.fmtobj:list}', expected_out_str)


class testFmt(unittest.TestCase):
    def setUp(self):
        self.data = ['The quick brown', 'Fox jumps over', 'The lazy dog']

    def test_disabled(self):
        fstr = psfmt(*self.data, disabled=True)
        self.assertEqual(fstr, self.data)

    def test_join(self):
        fstr = psfmt(*self.data, disabled=True, fmt_sep='\t')
        self.assertEqual(fstr, '\t'.join(self.data))

    def test_w_sep(self):
        fstr = psfmt(*self.data, fmt_sep='\n')
        self.assertEqual(fstr, '\n'.join(self.data))

    def test_wo_sep(self):
        fstr = psfmt(*self.data)
        self.assertEqual(fstr, self.data)

    def test_empty_sep(self):
        fstr = psfmt(*self.data, fmt_sep='')
        self.assertEqual(fstr, ''.join(self.data))



#!/usr/bin/env python3
# -*- coding:utf-8; mode:python -*-
#
# Copyright 2020, 2021 Pradyumna Paranjape
# This file is part of psprint.
#
# psprint is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# psprint is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with psprint.  If not, see <https://www.gnu.org/licenses/>.
#
'''
Prompt String-like Print
'''

import os
import sys

from .config import read_config
from .printer import PrintSpace


# Initiate default print function
def init_print(custom: os.PathLike = None) -> PrintSpace:
    '''
    Initiate ps-print function with default marks
    and marks read from various psprintrc configurarion files

    Args:
        custom: custom configuration file location

    '''

    default_print = read_config(custom)

    if 'idlelib.run' in sys.modules or not sys.stdout.isatty():
        # Running inside idle
        default_print.switches['bland'] = True

    return default_print


DEFAULT_PRINT = init_print()
'''
PrintSpace object created by reading defaults from various
psprintrc and psprint/style.yml files

'''

print = DEFAULT_PRINT.psprint
'''
psprint function for imports

'''

psfmt = DEFAULT_PRINT.psfmt
'''
ps formatting function for imports

'''

__all__ = ['DEFAULT_PRINT', 'print']

__version__ = '1!1.2.5'



#!/usr/bin/env python3
# -*- coding:utf-8; mode:python -*-
#
# Copyright 2020 Pradyumna Paranjape
# This file is part of psprint.
#
# psprint is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# psprint is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with psprint.  If not, see <https://www.gnu.org/licenses/>.
#
'''
Text Parts

'''

from typing import List, Tuple, Union

from .ansi import ANSI
from .errors import BadBGCol, BadColor, BadGloss, BadPrefix, BadShortPrefix


class AnsiEffect():
    '''
    Plain text object
    Text to be printed to (ANSI) terminal

    Args:
        color: color of text [0-15]
        gloss: gloss of text {0: bland, 1:normal ,2: dim, 3: bright}
        bgcol: color of background [0-15]

    Attributes:
        color: color of text
        gloss: gloss of text
        bgcol: background color

    Raises:
        BadColor
        BadGloss
        BadBGCol

    '''
    def __init__(self,
                 parent=None,
                 color: str = None,
                 gloss: str = None,
                 bgcol: str = None) -> None:
        # inherit
        self.color, self.gloss, self.bgcol = self.inherit(parent)

        # modify
        try:
            self.color: str = ANSI.FG_COLORS[color] if color else self.color
        except KeyError:
            raise BadColor(color) from None
        try:
            self.gloss: str = ANSI.GLOSS[gloss] if gloss else self.gloss
        except KeyError:
            raise BadGloss(gloss) from None
        try:
            self.bgcol: str = ANSI.BG_COLORS[bgcol] if bgcol else self.bgcol
        except KeyError:
            raise BadBGCol(bgcol) from None

    def __eq__(self, other) -> bool:
        if any(
                getattr(self, attr) != getattr(other, attr)
                for attr in ('color', 'gloss', 'bgcol')):
            return False
        return True

    def __str__(self) -> str:
        '''
        Human readable form
        '''
        return self.style

    def __repr__(self) -> str:  # pragma: no cover
        '''
        repr(self)
        '''
        return ' '.join((repr(self.color), repr(self.gloss), repr(self.bgcol)))

    @staticmethod
    def inherit(parent):
        '''
        object without parent
        set attributes to defaults
        '''
        if parent is None:
            return ANSI.FG_COLORS['t'], ANSI.GLOSS['n'], ANSI.BG_COLORS['t']
        return parent.color, parent.gloss, parent.bgcol

    @property
    def style(self) -> str:
        '''
        All style combined
        '''
        return self.color + self.bgcol + self.gloss


class PrintPref():
    '''
    Prefix that informs about Text

    Args:
        parent: template object for style
        pref: prefix in [long, short] format
        pref_max: pad with `space` to length
        **kwargs:
            * code:

                * color: {[0-15],[[l]krgybmcw],[[light] <color_name>]}
                * gloss: {[0-3],[rcdb],{reset,normal,dim,bright}}

            * for-

                * color: color of of prefix
                * gloss: gloss of prefix
                * bgcol: background color of prefix


    Arguments:
        pref: tuple: prefix long, short
        brackets: tuple: bool long, short
        pad: tuple: pad long, short
        style: AnsiEffect: color/gloss style

    Raises:
        BadPrefix
        BadShortPrefix

    '''
    def __init__(self,
                 parent=None,
                 pref: List[str] = None,
                 pref_max: int = 0,
                 **kwargs) -> None:
        self.brackets = [1, 1]
        style, self.pref = self._inherit(parent=parent, pref=pref)
        self.style = AnsiEffect(parent=style, **kwargs)
        # 0: long, 1: short
        pad_max = (pref_max, 1)
        pad_len: List[int] = [0, 0]
        for idx, pref_type in enumerate(self.pref):
            if not (isinstance(pref_type, str) or pref_type is None):
                raise (BadPrefix, BadShortPrefix)[idx](pref_type) from None
            if not pref_type:
                # pref_type is blank
                self.brackets[idx] = 0
                pad_len[idx] += 2  # corresponding to `[]`
            pad_len[idx] += max(pad_max[idx] - len(pref_type), 0)
        self.pad = [' ' * (span + 1) for span in pad_len]

    def __eq__(self, other) -> bool:
        if any(
                getattr(self, attr) != getattr(other, attr)
                for attr in ('pad', 'brackets', 'pref', 'style')):
            return False
        return True

    def __len__(self) -> int:  # pragma: no cover
        '''
        length of prefix
        '''
        return len(self.pref[0])

    @staticmethod
    def _inherit(
            parent=None,
            pref: List[str] = None) -> Tuple[Union[AnsiEffect, None], list]:
        '''
        inherit pref and style from parent if not supplied
        '''
        if parent is None:
            return None, pref or ['', '']
        if pref is None:  # pragma: no cover
            # Logically never reached unless explicitly stated
            pref = [parent.pref[0], parent.pref[1]]
        pref[0] = pref[0] or parent.pref[0]
        pref[1] = pref[1] or parent.pref[1]
        return parent.style, pref

    def to_str(self, **kwargs) -> str:
        '''
        Print prefix with style

        Args:
            short: prefix in short form?
            pad: Pad prefix
            bland: colorless pref
        '''
        pref_typ = int(kwargs.get('short', False))  # 1 if short, else 0
        parts = {
            'ansi': '' if kwargs.get('bland') else str(self.style),
            'text': self.pref[pref_typ],
            'brackets': self.brackets[pref_typ],
            'pref_pad': self.pad[pref_typ] if kwargs.get('pad') else '',
            'reset': '' if kwargs.get('bland') else ANSI.RESET_ALL,
        }
        return ''.join([
            parts['ansi'],
            '[' * parts['brackets'],
            parts['text'],
            ']' * parts['brackets'],
            parts['reset'],
            parts['pref_pad'],
        ])



#!/usr/bin/env python3
# -*- coding:utf-8; mode:python -*-
#
# Copyright 2020, 2021 Pradyumna Paranjape
# This file is part of psprint.
#
# psprint is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# psprint is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with psprint.  If not, see <https://www.gnu.org/licenses/>.
#
'''
module call
'''

from . import DEFAULT_PRINT, init_print

# Pass custom configuration here
CUSTOM_PRINT = init_print(custom=None)

# Some recipes are pre-initiated in psprint.__init__.py:
# psprint.DEFAULT_PRINT is initiated similarly as above
# psprint.print is psprint.DEFAULT_PRINT's psprint function
assert CUSTOM_PRINT == DEFAULT_PRINT

# Modify FLAGS
CUSTOM_PRINT.switches['short'] = False
CUSTOM_PRINT.switches['pad'] = True
# Standard flags should be specified in config file

# Add an InfoMark (style) 'in the script'
CUSTOM_PRINT.edit_style(pref='DOCS',
                        mark='docs',
                        pref_gloss='bright',
                        text_color='light blue')
# Multiple InfoMarks should be supplied through config file

# overload __builtin__.print to psprint.print
print = CUSTOM_PRINT.psprint
# If no customizations are intended, simply import psprint.print


def main():
    """
    main function call
    """
    # An iterable object
    edit_style_instructions = {
        'config': 'Pass custom config to psprint.init_print',
        'script': 'Edit psprint.DEFAULT_PRINT instance'
    }

    # Without any arguments, __builtin__.print() is called
    print()
    # Notice that blank lines don't have padding spaces

    # ``mark`` applies indicator styles
    print('usage:', mark='err')

    # Default mark is 'cont' i.e. no ``pref``, only padding
    print('Use me as an imported module')

    # Marks may be supplied as str, int or InfoMark objects
    print('Read usage instructions:', mark=1)

    # Using 'Non-shipped' mark, created on line 40
    print('https://pradyparanjpe.gitlab.io/psprint/usage.html', mark='docs')

    # 'shipped' marks are
    # 0: 'cont', 1: 'info', 2: 'act', 3: 'list', 4: 'warn', 5: 'err', 6: 'bug'
    print('from psprint import print', mark='act')
    print('from psprint import psfmt', mark='act')
    print('To edit styles, you may implement (any):', mark='info')

    # Iterable objects are resolved if instructed
    # Text may be indented (indentation level = 2 * <spaces>)
    print(edit_style_instructions, mark='act', indent=1, iterate=True)

    # __builtin__.print's kwargs are sincerely conveyed
    print('Capabilities are illustrated in',
          '``this``',
          'very file',
          mark='info',
          indent=2,
          sep=' ',
          end='\n')

    # ``pref`` and ``text`` style may be customized `on-the-fly`
    print(__file__, mark='docs', pref='FILE')

    # mark='bug' is a useful search query to sanitize debug code
    print('Bye', mark='bug')
    # print('User may ignore all lines marked with [DEBUG]', mark='bug')

    print()


if __name__ in ('__main__', 'psprint.__main__'):
    main()



#!/usr/bin/env python3
# -*- coding:utf-8; mode:python -*-
#
# Copyright 2020, 2021 Pradyumna Paranjape
# This file is part of psprint.
#
# psprint is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# psprint is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with psprint.  If not, see <https://www.gnu.org/licenses/>.
#
"""
Read configuration files from default locations
"""

import configparser
import os
import sys
from pathlib import Path
from typing import Dict, List

import toml
import yaml

from .errors import BadMark
from .printer import PrintSpace


def _is_mount(path: Path):
    """
    Check across platform if path is mountpoint or drive

    Args:
        path: path to be checked
    """
    try:
        if path.is_mount():
            return True
        return False
    except NotImplementedError:
        if path.resolve().drive + '\\' == str(path):
            return True
        return False


def _parse_yaml(config: Path) -> Dict[str, dict]:
    """
    Read configuration specified as a yaml file:
        - .psprintrc
        - style.yml
        - *.yml
    """
    with open(config, 'r') as rcfile:
        conf: Dict[str, dict] = yaml.safe_load(rcfile)
    if conf is None:  # pragma: no cover
        raise yaml.YAMLError
    return conf


def _parse_ini(config: Path, sub_section: bool = False) -> Dict[str, dict]:
    """
    Read configuration supplied in ``setup.cfg`` OR
        - *.cfg
        - *.conf
        - *.ini
    """
    parser = configparser.ConfigParser()
    parser.read(config)
    if sub_section:
        return {
            pspcfg.replace('psprint.', ''): dict(parser.items(pspcfg))
            for pspcfg in parser.sections() if 'psprint' in pspcfg
        }
    return {
        pspcfg: dict(parser.items(pspcfg))
        for pspcfg in parser.sections()
    }  # pragma: no cover


def _parse_toml(config: Path, sub_section: bool = False) -> Dict[str, dict]:
    """
    Read configuration supplied in ``pyproject.toml`` OR
        - *.toml
    """
    if sub_section:
        with open(config, 'r') as rcfile:
            conf: Dict[str, dict] = toml.load(rcfile).get('psprint', {})
        return conf
    with open(config, 'r') as rcfile:
        conf: Dict[str, dict] = dict(toml.load(rcfile))
    if conf is None:  # pragma: no cover
        raise toml.TomlDecodeError
    return conf


def _parse_rc(config: Path) -> Dict[str, dict]:
    """
    Parse psprintrc file

    Args:
        config: path to configuration file

    Returns:
        configuration sections

    Raises:
        BadMark: Bad configuration

    """
    if config.name == 'setup.cfg':
        # declared inside setup.cfg
        return _parse_ini(config, sub_section=True)
    if config.name == 'pyproject.toml':
        # declared inside pyproject.toml
        return _parse_toml(config, sub_section=True)
    try:
        # yaml configuration format
        return _parse_yaml(config)
    except yaml.YAMLError:
        try:
            # toml configuration format
            return _parse_toml(config)
        except toml.TomlDecodeError:
            try:
                # try generic config-parser
                return _parse_ini(config)
            except configparser.Error:
                raise BadMark(mark='configuration', config=str(config))


def read_config(custom: os.PathLike = None) -> PrintSpace:
    """
    Read psprint configurations from various locations

    Args:
        custom: custom location for configuration

    Returns:
        configured ``PrintSpace``

    Raises:
        BadMark- Bad configuration file format
    """

    # environment
    if sys.platform.startswith('win'):  # pragma: no cover
        # windows
        user_home = Path(os.environ['USERPROFILE'])
        root_config = Path(os.environ['APPDATA'])
        xdg_config = Path(
            os.environ.get('LOCALAPPDATA', user_home / 'AppData/Local'))
    else:
        # assume POSIX
        user_home = Path(os.environ['HOME'])
        root_config = Path('/etc')
        xdg_config = Path(
            os.environ.get('XDG_CONFIG_HOME', user_home / '.config'))

    # Preference of configurations *Most dominant first*
    config_heir: List[Path] = []

    # custom
    if custom is not None:
        if not Path(custom).is_file():
            raise FileNotFoundError(
                f'Custom configuration file: {custom} not found')
        config_heir.append(Path(custom))

    # environment variable
    psprintrc_var = os.environ.get('PSPRINTRC')
    if psprintrc_var is not None:
        if not Path(psprintrc_var).is_file():
            raise FileNotFoundError(
                f'PSPRINTRC configuration file: {psprintrc_var} not found')
        config_heir.append(Path(psprintrc_var))

    # current- and all parent- directories till project root or mountpoint
    # project root is the directory, which contains setup.(cfg|py) file
    current_dir = Path('.').resolve()
    while not (_is_mount(current_dir)):
        if current_dir / '__init__.py':
            config_heir.append((current_dir / '.psprintrc'))
        if any((current_dir / setup).is_file()
               for setup in ('setup.cfg', 'setup.py')):
            config_heir.append((current_dir / 'pyproject.toml'))
            config_heir.append((current_dir / 'setup.cfg'))
            break
        current_dir = current_dir.parent

    # standard locations
    for heir in xdg_config, root_config, Path(__file__).parent.parent:
        for ext in '.yml', '.yaml', '.toml', '.conf':
            config_heir.append((heir / 'psprint/style').with_suffix(ext))

    # load configs from oldest ancestor to current directory
    default_print = PrintSpace()
    for config in reversed(config_heir):
        try:
            default_print.set_opts(_parse_rc(config), str(config))
        except (PermissionError, FileNotFoundError, IsADirectoryError):
            pass

    # initialize with config

    return default_print



#!/usr/bin/env python3
# -*- coding:utf-8; mode:python -*-
#
# Copyright 2020 Pradyumna Paranjape
# This file is part of psprint.
#
# psprint is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# psprint is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with psprint.  If not, see <https://www.gnu.org/licenses/>.
#
'''
Information- Prepended Print object
'''

from collections.abc import Iterable
from sys import stderr, stdout
from types import GeneratorType
from typing import Any, Dict, Generator, List, Optional, Tuple, Union
from warnings import warn

from .ansi import ANSI
from .errors import BadMark, DeepIterWarning, LongIterWarning
from .mark_types import InfoMark


class PrintSpace():
    """
    Fancy Print class that also prints the type of message

    Args:
        conf: parsed configuration
        name: name of configuration file (to report errors)

    Attributes:
        pref_max: int: maximum length of prefix string
        info_style; dict: pre-configured prefix styles
        info_index: list: keys of `info_style` mapped to int
        print_kwargs: dict : library of kwargs_accepted by print_function
        switches: dict: user-customizations: pad, short, bland, disabled

            pad: bool: prefix is padded to start text at the same level
            short: bool: display short, 1 character- prefix
            bland: bool: do not show ANSI color/styles for prefix/text
            disabled: bool: behave like python default print_function

    """
    def __init__(self, conf: Dict[str, dict] = None, name: str = None) -> None:
        # Standard info styles
        self.switches = {
            'pad': False,
            'short': False,
            'bland': False,
            'iterate': False,
            'disabled': False
        }
        self.print_kwargs = {
            'file': stdout,
            'sep': "\t",
            'end': "\n",
            'flush': False
        }
        self.pref_max = None
        self.info_style: Dict[str, InfoMark] = {}
        self.info_index: List[str] = []
        if conf is not None:
            self.set_opts(conf, name or 'unnamed config file')

    def __repr__(self) -> str:
        """
        Returns:
            Formatted summary of info_style
        """
        outstr = '\nid\tmark\tlong\tshort\ttext\n\n'
        outstr += "\n".join((f'{i}' + '\t' + f'{k}:{v}' for i, k, v in self))
        return outstr

    def __contains__(self, mark: Union[str, int, InfoMark] = None) -> bool:
        """
        is mark defined?
        """
        if mark is None:
            return False
        if isinstance(mark, str):
            return mark in self.info_style.keys()
        if isinstance(mark, int):
            return len(self.info_index) > mark
        return mark in self.info_style.values()

    def __getitem__(self, mark: Union[int, str]) -> InfoMark:
        """
        Returns:
            Infomark corresponding to mark
        """
        try:
            # check if mark can be converted to int index
            mark = int(mark)
        except ValueError:
            if isinstance(mark, str):
                return self.info_style.get(mark, self.info_style['cont'])
        except Exception:
            raise KeyError(mark)
        if 0 <= mark < len(self.info_index):
            return self.info_style[self.info_index[mark]]
        return self.info_style[self.info_index[0]]

    def __iter__(self) -> Generator[Tuple[int, str, InfoMark], None, None]:
        for index_int, mark in enumerate(self.info_index):
            yield index_int, mark, self.info_style[mark]
        return

    def __eq__(self, value: 'PrintSpace') -> bool:
        """
        Equate
        """
        if any(
                getattr(self, child) != getattr(value, child)
                for child in ('pref_max', 'info_style', 'info_index',
                              'print_kwargs', 'switches')):
            return False
        return True

    def set_opts(self, conf: Dict[str, dict], name: str) -> None:
        """
        configuration from rcfile

        Args:
            conf: parsed configuration
            name: name of configuration file (to report errors)

        Raises:
            BadMark

        """
        if not conf:
            return
        info_index: Optional[Dict[str, str]] = None
        for mark, settings in conf.items():
            if mark == "FLAGS":
                # switches / flags
                self.pref_max = settings.get("pref_max", None)

                # update psprint switches
                self.switches.update(
                    dict(
                        filter(lambda x: x[0] in self.switches,
                               settings.items())))

                # update __builtin__,print kwargs
                self.print_kwargs.update(
                    dict(filter(lambda x: x[0] in self.print_kwargs,
                                settings)))

                # standard output file
                fname = settings.get("file", None)  # Discouraged
                if fname is not None:  # pragma: no cover
                    self.print_kwargs['file'] = open(fname, "a")

            elif mark == 'order':
                info_index = settings
            else:
                # New mark definition
                try:
                    self.edit_style(mark=mark, **settings)
                except (ValueError, TypeError):  # pragma: no cover
                    raise BadMark(str(mark), name) from None
        if info_index is not None:
            self.info_index = list(
                filter(lambda x: x in self.info_style, info_index))

    def edit_style(self,
                   pref: str,
                   index_int: int = None,
                   mark: str = None,
                   **kwargs) -> str:
        """
        Edit loaded style

        Args:
            pref: str: prefix string long [length < 10 characters]
            index_int: Index number that will call this InfoMark
            mark: Mark string that will call this ``InfoMark``
            **kwargs:
                * pref_s: str: prefix string short [1 character]
                * code:

                    * color: {[0-15],[[l]krgybmcw],[[light] <color_name>]}
                    * gloss: {[0-3],[rcdb],{reset,normal,dim,bright}}

                * for-

                    * pref_color: color of of prefix
                    * pref_gloss: gloss of prefix
                    * pref_bgcol: background color of prefix
                    * text_color: color of of text
                    * text_gloss: gloss of text
                    * text_bgcol: background color of text

        Returns
            Summary of new (updated) ``PrintSpace``

        """
        # correct pref
        if mark is None:
            mark = pref[:4]
        kwargs['pref'] = pref
        if index_int is None or \
           not 0 <= index_int <= len(self.info_index):
            self.info_index.append(mark)
        else:
            self.info_index.insert(index_int, mark)
        self.info_style[mark] = InfoMark(pref_max=self.pref_max, **kwargs)
        return str(self)

    def remove_style(self, mark: Union[str, int, InfoMark]) -> str:
        """
        Args:
            mark: is popped out of defined styles
            index_int: is used to locate index_str if it is not provided

        Returns
            Summary of new (updated) ``PrintSpace``

        """
        _idx = None
        _mark = None
        if mark not in self:
            raise KeyError(mark)
        if isinstance(mark, int):
            _idx = mark
            _mark = self.info_index[mark]
        elif isinstance(mark, str):
            _idx = self.info_index.index(mark)
            _mark = mark
        elif isinstance(mark, InfoMark):
            for idx, t_mark, info_mark in self:
                if info_mark == mark:
                    _idx = idx
                    _mark = t_mark
                    break
        if _idx is None or _mark is None:
            # This should never happen
            raise KeyError(mark)  # pragma: no cover
        del self.info_style[_mark]
        del self.info_index[_idx]
        return repr(self)

    def _which_mark(self,
                    mark: Union[str, int, InfoMark] = None,
                    **kwargs) -> InfoMark:
        """
        Define a mark based on arguments supplied

            * may be a pre-defined mark OR
            * mark defined on the fly

        Args:
            mark: mark that identifies a defined prefix
            **kwargs:
                * pref: str: prefix string long [length < 10 characters]
                * pref_s: str: prefix string short [1 character]
                * code:

                    * color: {[0-15],[[l]krgybmcw],[[light] <color_name>]}
                    * gloss: {[0-3],[rcdb],{reset,normal,dim,bright}}

                * for-

                    * pref_color: color of of prefix
                    * pref_gloss: gloss of prefix
                    * pref_bgcol: background color of prefix
                    * text_color: color of of text
                    * text_gloss: gloss of text
                    * text_bgcol: background color of text

        """
        base_mark: InfoMark = self.info_style['cont']
        if mark is not None:
            # mark was supplied
            if isinstance(mark, InfoMark):
                # ready-made mark was served
                base_mark = mark
            elif isinstance(mark, (str, int)):
                # mark supplied as index int
                base_mark = self[mark]
            else:
                raise BadMark(mark=str(mark), config="**kwargs")
        if any(arg in kwargs for arg in [
                'pref',
                'pref_s',
                'pref_color',
                'pref_gloss',
                'pref_bgcol',
                'text_color',
                'text_gloss',
                'text_bgcol',
        ]):
            return InfoMark(parent=base_mark, pref_max=self.pref_max, **kwargs)
        return base_mark

    def _iter_print(self,
                    args: Any,
                    mark: Union[str, int, InfoMark] = None,
                    **kwargs):
        # sub-section head
        arg_type = type(args)
        sub_head = {
            list: 'list:',
            dict: 'dict:',
            tuple: 'tuple:',
            GeneratorType: 'generator:'
        }.get(arg_type, str(arg_type))
        self.psprint(sub_head, mark=mark, **kwargs)

        # Check if args can be easily converted to a dict
        if isinstance(args, GeneratorType):
            args = {idx: val for idx, val in enumerate(args)}
        else:
            try:
                # generator consumes an argument here
                args = dict(args)
            except (TypeError, ValueError):
                # create dictionary of args
                args = {idx: val for idx, val in enumerate(args)}

        # Alter kwargs
        if 'pref' in kwargs:
            del kwargs['pref']
        if 'end' in kwargs:
            del kwargs['end']
        kwargs['indent'] = kwargs.get('indent', 0) + 1
        if kwargs['indent'] > 10:
            self.psprint('...deep stack...', mark='warn', **kwargs)
            warn('...deep stack...', category=DeepIterWarning)
            return

        _num_elements = 0
        for key, value in args.items():
            _num_elements += 1
            if _num_elements > 100:
                self.psprint('...long stack...',
                             mark='warn',
                             **kwargs,
                             file=stderr)
                warn('...long stack...', category=LongIterWarning)
                return
            if isinstance(value, Iterable) and not isinstance(value, str):
                self.psprint(f'{key}:', mark='list', end='', **kwargs)
                self.psprint(value, **kwargs)
            else:
                self.psprint(f'{key}: {value}',
                             mark='list',
                             **kwargs,
                             file=stderr)

    def psfmt(self,
              *args,
              mark: Union[str, int, InfoMark] = None,
              fmt_sep: str = None,
              **kwargs) -> Union[List[str], str]:
        """
        Prefix String represenattion.

        Args:
            *args: passed to print_function for printing
            mark: pre-configured `InfoMark` defaults:

                * cont or 0 or anything else: nothing
                * info or 1: [INFO]
                * act  or 2: [ACTION]
                * list or 3: [LIST]
                * warn or 4: [WARNING]
                * error:or 5: [ERROR]
                * bug:  or 6 [DEBUG]
                * `Other marks defined in .psprintrc`

            fmt_sep: If not ``None``, return `*args` joined by separator.

            **kwargs:
                * pref: str: prefix string long [length < 10 characters]
                * pref_s: str: prefix string short [1 character]
                * code:

                    * color: {[0-15],[[l]krgybmcw],[[light] <color_name>]}
                    * gloss: {[0-3],[rcdb],{reset,normal,dim,bright}}

                * for-

                    * pref_color: color of of prefix
                    * pref_gloss: gloss of prefix
                    * pref_bgcol: background color of prefix
                    * text_color: color of of text
                    * text_gloss: gloss of text
                    * text_bgcol: background color of text

                * pad: bool: prefix is padded to start text at the same level
                * short: bool: display short, 1 character- prefix
                * bland: bool: do not show ANSI color/styles for prefix/text
                * disabled: bool: behave like python default print_function
                * indent: indent level of print statement (1 indent =2 spaces)

        Raises:

            BadMark: mark couldn't be interpreted

        Returns:

            * PSPRINT-like represented args. When `these` args is printed
              using standard print, PSPRINT-like output appears.
            * If a fmt_sep is provided,
              it is used to join args and return a string

        """
        switches = {
            key: {
                **self.switches,
                **kwargs
            }[key]
            for key in self.switches
        }

        # typecast to list[str]
        args_l = list(str(arg) for arg in args)
        if switches['disabled'] or not args_l:
            if fmt_sep is not None:
                return fmt_sep.join(args_l)
            return args_l

        mark = self._which_mark(mark=mark, **kwargs)

        # add indentation
        args_l[0] = '  ' * max(int(kwargs.get('indent', 0)), 0) + args_l[0]

        # add prefix to *args[0]
        if not switches.get('bland'):  # pragma: no cover
            args_l[0] = str(mark.text) + str(args_l[0])
            args_l[-1] = str(args_l[-1]) + ANSI.RESET_ALL
        args_l[0] = mark.pref.to_str(**switches) + str(args_l[0])
        if fmt_sep is not None:
            return fmt_sep.join(args_l)
        return args_l

    def psprint(self,
                *args,
                mark: Union[str, int, InfoMark] = None,
                **kwargs) -> None:
        """
        Prefix String PRINT

        Args:
            *args: passed to print_function for printing
            mark: pre-configured `InfoMark` defaults:

                * cont or 0 or anything else: nothing
                * info or 1: [INFO]
                * act  or 2: [ACTION]
                * list or 3: [LIST]
                * warn or 4: [WARNING]
                * error:or 5: [ERROR]
                * bug:  or 6 [DEBUG]
                * `Other marks defined in .psprintrc`

            **kwargs:
                * pref: str: prefix string long [length < ``pref_max``]
                * pref_s: str: prefix string short [1 character]
                * code:

                    * color: {[0-15],[[l]krgybmcw],[[light] <color_name>]}
                    * gloss: {[0-3],[rcdb],{reset,normal,dim,bright}}

                * for-

                    * pref_color: color of of prefix
                    * pref_gloss: gloss of prefix
                    * pref_bgcol: background color of prefix
                    * text_color: color of of text
                    * text_gloss: gloss of text
                    * text_bgcol: background color of text

                * pad: bool: prefix is padded to start text at the same level
                * short: bool: display short, 1 character- prefix
                * bland: bool: do not show ANSI color/styles for prefix/text
                * disabled: bool: behave like python default print_function
                * indent: int: indent level of print statement (1 indent =2 spaces)
                * iterate: int: Resolve recursively if

                  - only 1 ``*args`` is supplied,
                  - it is iterable
                  - it is not ``str``

                * file: IO: passed to print function
                * sep: str: passed to print function
                * end: str: passed to print function
                * flush: bool: passed to print function

        Raises:
            BadMark: mark couldn't be interpreted

        """
        # Extract print-kwargs
        print_kwargs = {
            key: {
                **self.print_kwargs,
                **kwargs
            }[key]
            for key in self.print_kwargs
        }

        if len(args) == 1:
            if all(condition for condition in (
                    kwargs.get('iterate', self.switches['iterate']),
                    isinstance(args[0], Iterable),
                    not isinstance(args[0], str))):
                self._iter_print(args[0], mark=mark, **kwargs)
                return

        print(*self.psfmt(*args, mark=mark, **kwargs), **print_kwargs)



#!/usr/bin/env python3
# -*- coding:utf-8; mode:python -*-
#
# Copyright 2020, 2021 Pradyumna Paranjape
# This file is part of psprint.
#
# psprint is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# psprint is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with psprint.  If not, see <https://www.gnu.org/licenses/>.
#
'''
test print
'''

import tempfile
import unittest
from pathlib import Path

from psprint import DEFAULT_PRINT, errors, init_print
from psprint.mark_types import InfoMark
from psprint.printer import PrintSpace


class TestPrintSpace(unittest.TestCase):
    """
    Test basic functions
    """
    def setUp(self):
        self.test_print = init_print()

    def test_eq(self):
        self.assertEqual(self.test_print, self.test_print)

    def test_ne(self):
        mod_print = init_print()
        mod_print.edit_style('test', mark='test')
        self.assertNotEqual(self.test_print, mod_print)

    def test_defaults_load(self):
        """
        print all default marks
        """
        print(self.test_print)

    def test_eq_compare_marks(self):
        """
        compare InfoMarks
        """
        self.assertEqual(self.test_print[0], self.test_print[0])

    def test_eq_compare_ansi(self):
        """
        compare text of marks
        """
        self.assertEqual(self.test_print[0].text, self.test_print[0].text)

    def test_ne_compare_marks(self):
        """
        compare InfoMarks
        """
        self.assertNotEqual(self.test_print[0], self.test_print[1])

    def test_ne_compare_ansi(self):
        """
        compare text of marks
        """
        self.assertNotEqual(self.test_print[6].text, self.test_print[1].text)

    def test_disabled(self):
        """
        Disable fancy
        """
        self.test_print.psprint('Disabled', disabled=True)

    def test_default_mark(self):
        """
        Default print
        """
        self.test_print.psprint("Info Mark", mark='info')

    def test_new_mark(self):
        """
        Test kwarg
        """
        configuration = '\n'.join([
            'FLAGS:',
            '  short: No',
            '  pad: Yes',
            '',
            'TEST:',
            '  pref: TEST',
            '  pref_s: t',
            '  pref_color: y',
            '  pref_bgcol: blue',
            '  pref_gloss: 1',
            '  text_color: k',
            '  text_bgcol: 7',
            '  text_gloss: d',
            '',
        ])
        with tempfile.NamedTemporaryFile('wt', delete=False) as config:
            config.write(configuration)
        my_print = init_print(Path(config.name))
        my_print.psprint("Test Text", mark='TEST')
        my_print.psprint("Info Test", mark='info')
        my_print.set_opts({}, '')
        config.close()
        my_print.remove_style('TEST')
        PrintSpace({})

    def test_not_in(self):
        self.assertNotIn('badkey', self.test_print)

    def test_mark_in(self):
        self.assertIn(self.test_print.info_style['info'], self.test_print)
        self.assertIn('info', self.test_print)

    def test_on_the_fly(self):
        """
        Test definition on the fly
        """
        self.test_print.psprint("OTF text",
                                pref="OTF",
                                pref_s="o",
                                text_bgcol='lg',
                                pref_color='r',
                                short=True)

    def test_mod_mark(self):
        """
        Test a modified mark
        """
        self.test_print.psprint("MOD MARK test",
                                mark=2,
                                pref_color="lg",
                                pref=None,
                                pref_s=None)

    def test_edit(self):
        """
        test edit style
        """
        self.test_print.edit_style(pref="my_test", index_int=4)

    def test_int_pref(self):
        """
        int/float as prefix
        """
        self.test_print.psprint("bad prefix", pref=1234)

    def test_pop_style(self):
        """
        test edit style
        """
        self.test_print.remove_style(4)

    def test_remove_style_mark(self):
        """
        test removal by passing InfoMark object
        """
        rem_mark = self.test_print[0]
        print(rem_mark)
        self.test_print.remove_style(rem_mark)

    def test_allow_unknown_mark(self):
        """
        allow an unknown mark (interpret as 0/cont)
        """
        self.test_print.psprint("Test Unknown", mark=88)

    def test_precreated_mark(self):
        mark = InfoMark(pref='prepref')
        self.test_print.psprint("pass mark", mark=mark)

    def test_precreated_inherit(self):
        mark = InfoMark(parent=self.test_print.info_style['info'])
        self.test_print.psprint("pass mark", mark=mark)


class TextErrors(unittest.TestCase):
    def test_warns(self):
        """
        Check that warnings are thrown
        """
        self.assertWarns(errors.PSPrintWarning,
                         DEFAULT_PRINT.psprint,
                         "warn_text",
                         mark=5,
                         pref="SOME LONG TEXT",
                         pref_s="LONG SHORT")

    def test_bad_pref(self):
        """
        trigger a bad prefix
        """
        self.assertRaises((errors.BadPrefix, errors.BadShortPrefix),
                          DEFAULT_PRINT.psprint,
                          "bad prefix",
                          pref=["bad list prefix"])

    def test_bad_col(self):
        """
        trigger a bad color error
        """
        self.assertRaises(errors.BadColor,
                          DEFAULT_PRINT.psprint,
                          "bad color",
                          pref_color=77)
        self.assertRaises(errors.BadBGCol,
                          DEFAULT_PRINT.psprint,
                          "bad color",
                          pref_bgcol=77)
        self.assertRaises(errors.BadGloss,
                          DEFAULT_PRINT.psprint,
                          "bad gloss",
                          pref_gloss=77)

    def test_bad_rm_mark(self):
        """
        test mark removal error
        """
        self.assertRaises(KeyError, DEFAULT_PRINT.remove_style, None)

    def test_bad_mark(self):
        """
        test bad mark creation
        """
        self.assertRaises(errors.BadMark,
                          DEFAULT_PRINT.psprint,
                          "bad mark",
                          mark=[])

    def test_no_file(self):
        """
        custom-supplied file missing
        """
        self.assertRaises(FileNotFoundError, init_print, "bad_file_name")

    def test_nokey(self):
        """
        bad key not in PrintSpace
        """
        self.assertRaises(KeyError, DEFAULT_PRINT.__getitem__, list())


class TestIter(unittest.TestCase):
    @staticmethod
    def test_iter_list():
        DEFAULT_PRINT.psprint([1, [2, 20, 200], 3, 4], mark="info")

    @staticmethod
    def test_iter_gen():
        DEFAULT_PRINT.psprint(
            (k for k in (1, (t for t in (2, 20, 200)), 3, 4)), mark="info")

    @staticmethod
    def test_iter_dict():
        DEFAULT_PRINT.psprint(
            {
                "a": 1,
                "b": 2,
                3: {
                    "int": 3,
                    "bool": bin(3),
                    "hex": hex(3)
                }
            },
            mark="info")

    @staticmethod
    def test_iter_all():
        DEFAULT_PRINT.psprint((k for k in [[], {
            "a": 1,
            "b": [2, "b"],
            3: {
                "int": 3,
                "bool": bin(3),
                "hex": hex(3)
            }
        }, {1, 2, 3, 4}]),
                              mark="info")



#!/usr/bin/env python3
# -*- coding:utf-8; mode:python -*-
#
# Copyright 2020 Pradyumna Paranjape
# This file is part of psprint.
#
# psprint is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# psprint is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details. #
# You should have received a copy of the GNU Lesser General Public License
# along with psprint.  If not, see <https://www.gnu.org/licenses/>.
#
'''
Information Marker

'''

import warnings
from typing import Dict, Union

from .ansi import ANSI
from .errors import ValueWarning
from .text_types import AnsiEffect, PrintPref

DEFAULT_STYLE: Dict[str, int] = {'color': 16, 'gloss': 1, 'bgcol': 16}
'''
Terminal-determined color, black background, normal gloss

'''


class InfoMark():
    '''
    Prefix Mark information

    Attributes:
        pref: PrintPref: Prefix text properties
        text: PrintText: Text properties

    Args:
        parent: Inherit information from-
        pref_max: pad prefix to reach length
        **kwargs:
            * pref: prefix string
            * pref_s: prefix short string
            * code:

                * color: {[0-15],[[l]krgybmcw],[[light] <color_name>]}
                * gloss: {[0-3],[rcdb],{reset,normal,dim,bright}}

            * for-

                * pref_color: color of of prefix
                * pref_gloss: gloss of prefix
                * pref_bgcol: background color of prefix
                * text_color: color of of text
                * text_gloss: gloss of text
                * text_bgcol: background color of text

    '''
    def __init__(self,
                 parent: 'InfoMark' = None,
                 pref_max: int = None,
                 **kwargs: Union[str, str]) -> None:
        if pref_max is None:
            pref_max = 7  # default

        # categorise kwargs
        pref_args = {}
        for key in DEFAULT_STYLE:
            if f'pref_{key}' in kwargs:
                pref_args[key] = kwargs[f'pref_{key}']
        text_args = {}
        for key in DEFAULT_STYLE:
            if f'text_{key}' in kwargs:
                text_args[key] = kwargs[f'text_{key}']

        # Determine prefixes
        for p_type in 'pref', 'pref_s':
            if kwargs.get(p_type) is not None:
                if isinstance(kwargs[p_type], (int, float)):
                    kwargs[p_type] = str(kwargs[p_type])
        pref = [kwargs.get('pref', ''), kwargs.get('pref_s', '>')]
        if pref[0] is not None and len(pref[0]) > pref_max:
            trim = pref[0][:pref_max]
            warnings.warn(f"Prefix string '{pref[0]}'" +
                          f" is too long (length>{pref_max}) " +
                          f"trimming to {trim}",
                          category=ValueWarning)
            pref[0] = trim
        if pref[1] is not None and len(pref[1]) > 1:
            trim = pref[1][:1]
            warnings.warn("Prefix string '{pref[1]}'" +
                          f" is too long (length>1) trimming to {trim}",
                          category=ValueWarning)
            pref[1] = trim
        parent_pref = parent.pref if parent else None
        parent_text = parent.text if parent else None

        self.pref = PrintPref(parent=parent_pref,
                              pref=pref,
                              pref_max=pref_max,
                              **pref_args)
        self.text = AnsiEffect(parent=parent_text, **text_args)

    def __repr__(self) -> str:
        '''
        String format of available information
        '''
        return "\t".join(
            (str(self.pref.style), self.pref.pref[0], self.pref.pref[1],
             self.text.style + "<CUSTOM>" + ANSI.RESET_ALL))

    def __eq__(self, other) -> bool:
        if any(
                getattr(self, attr) != getattr(other, attr)
                for attr in ('pref', 'text')):
            return False
        return True

    def get_info(self) -> str:  # pragma: no cover
        '''
        Print information about ``InfoMark``

        '''
        info = str(self)
        print(info)
        return info
